/*=========================================================================

 Program:   vv
 Module:    $RCSfile: vvSlicersManager.h,v $
 Language:  C++
 Date:      $Date: 2008/07/15 11:26:35 $
 Version:   $Revision: 1.17 $
 Author :   Pierre Seroul (pierre.seroul@gmail.com)

Copyright (C) 2008
Léon Bérard cancer center http://oncora1.lyon.fnclcc.fr
CREATIS-LRMN http://www.creatis.insa-lyon.fr

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=========================================================================*/
#ifndef vvSlicersManager_h
#define vvSlicersManager_h

// std
#include <iostream>
#include <vector>

// QT
#include <QObject>

// VTK
#include <vtkImageData.h>
#include <vtksys/SystemTools.hxx>

// VV
#include "vvSlicer.h"
#include "vvImage.h"
#include "vvInteractorStyleNavigator.h"
#include "vvConvertITKImageToVTKDataFilter.h"
#include "vvConvertITKVectorToVTKDataFilter.h"
#include "vvLandmarks.h"
#include "vvCommon.h"

class vvSlicersManager : public QObject {
  Q_OBJECT

public:
  vvSlicersManager(int numberOfSlicers);
  ~vvSlicersManager();

  std::string GetLastError(){return mLastError;}

  template<class ImageType>
  bool SetImageFromITK(std::string filename, typename ImageType::Pointer image);

  bool SetImage(std::string filename,int dim, std::string component,int type);
  void SetExtractedImage(std::string filename, vvImage* image, int dim, std::string component, int slice);
  bool SetImages(std::vector<std::string> filenames,int dim, std::string component,int type);

  bool SetOverlay(std::string filename, int dim, std::string component);
  bool SetFusion(std::string filename, int dim, std::string component);
  bool SetVF(std::string filename,int dim, std::string component);

  std::string GetFileName(){return mFileName;}
  std::string GetVFName(){return mVFName;}
  std::string GetOverlayName(){return mOverlayName;}
  std::string GetFusionName(){return mFusionName;}

  vvSlicer* GetSlicer(int i);
  void UpdateSlicer(int num, int state);
  void SetSlicerWindow(int i, vtkRenderWindow* RW);
  void SetInteractorStyleNavigator(int i,vvInteractorStyleNavigator* style);
  int NumberOfSlicers() {return mSlicers.size();}

  std::string GetComponent(){return mComponent;}
  std::string GetVFComponent(){return mVFComponent;}
  int GetDimension(){return mDim;}
  int GetType(){return mType;}

  void SetId(std::string id){mId = id;}
  std::string GetId(){return mId;}

  void SetFilename(std::string f) { mFileName = f; }

  void SetTSlice(int slice);
  void SetNextTSlice();
  void SetPreviousTSlice();
  void SetTSliceInSlicer(int tslice, int slicer);

  void GenerateDefaultLookupTable();
  void SetColorWindow(double s);
  void SetColorLevel(double s);
  void SetOpacity(int i, double factor);
  void SetColorMap(int colormap);
  void SetPreset(int preset);
  void SetOverlayColor(int color){mOverlayColor = color;}
  void SetFusionOpacity(int opacity){mFusionOpacity = opacity;}
  void SetFusionColorMap(int colorMap){mFusionColorMap = colorMap;}
  void SetFusionWindow(int window){mFusionWindow = window;}
  void SetFusionLevel(int level){mFusionLevel = level;}

  double GetColorWindow();
  double GetColorLevel();
  int GetColorMap(){return mColorMap;}
  int GetPreset(){return mPreset;}
  int GetOverlayColor(){return mOverlayColor;}

  int GetFusionOpacity(){return mFusionOpacity;}
  int GetFusionColorMap(){return mFusionColorMap;}
  double GetFusionWindow(){return mFusionWindow;}
  double GetFusionLevel(){return mFusionLevel;}

  void SetCursorVisibility(int s);
  void ResetWindowLevel();
  void UpdateViews(int current, int slicer);
  void UpdateLinked(int slicer);
  void Render();

  void AddLink(std::string newId){mLinkedId.push_back(newId);}
  void RemoveLink(std::string oldId){mLinkedId.remove(oldId);}

  void RemoveVF();
  void RemoveOverlay();
  void RemoveFusion();
  void RemoveActors();
  void Reload();
  void ReloadVF();

  void Activated();
  void UpdateInfoOnCursorPosition(int slicer);
  void UpdateWindowLevel();
  void UpdateSlice(int slicer);
  void UpdateTSlice(int slicer);
  void UpdateSliceRange(int slicer);

  vvLandmarks *GetLandmarks();
  void AddLandmark(float x,float y,float z,float t);

signals :
  void currentImageChanged(std::string id);
  void UpdatePosition(int visibility,double x, double y, double z, double X, double Y, double Z, double value);
  void UpdateVector(int display, double x, double y, double z, double value);
  void UpdateOverlay(int display, double valueOver, double valueRef);
  void UpdateFusion(int display, double valueFus);
  void UpdateWindows(int slicer, int view, int slice);
  void UpdateSlice(int slicer, int slice);
  void UpdateTSlice(int slicer, int slice);
  void UpdateSliceRange(int slice, int min, int max, int tmin, int tmax);
  void WindowLevelChanged(double window, double level, int preset, int colormap);
  void UpdateLinkManager(std::string, int slicer, double x, double y, double z, int temps);
  void LandmarkAdded();

private:
  std::vector<vvSlicer*> mSlicers;
  vvConvertITKImageToVTKDataFilter* mConvertor;
  vvConvertITKImageToVTKDataFilter* mOverlayConvertor;
  vvConvertITKImageToVTKDataFilter* mFusionConvertor;
  vvConvertITKVectorToVTKDataFilter* mVectorConvertor;
  int mDim;
  int mColorMap;
  int mOverlayColor;

  int mFusionOpacity;
  int mFusionColorMap;
  double mFusionWindow;
  double mFusionLevel;

  int mPreset;
  int mType;
  std::string mComponent;
  std::string mVFComponent;
  std::string mOverlayComponent;
  std::string mFusionComponent;
  std::string mFileName;
  std::string mId;
  std::string mVFName;
  std::string mOverlayName;
  std::string mFusionName;
  std::string mVFId;
  std::string mLastError;
  std::list<std::string> mLinkedId;

  vvLandmarks* mLandmarks;
};

#include "vvSlicersManager.txx"

#endif
